/* ------------------------------------------------------------------------- */
/*
 *  true_type.h
 *
 *  Copyright (c) 2009 - 2010, clown. All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    - No names of its contributors may be used to endorse or promote
 *      products derived from this software without specific prior written
 *      permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  Copyright (c) 2000,2001,2002,2003,2004,2005,2006,2007,2008,2009
 *  by George Williams 
 *
 *  The source code is derived from the FontForge Project. (See the
 *  following URL: http://fontforge.sourceforge.net/)
 *
 *  Last-modified: Thu 07 Jan 2010 16:49:00 JST
 */
/* ------------------------------------------------------------------------- */
#ifndef CLIFF_FONT_TRUE_TYPE_H
#define CLIFF_FONT_TRUE_TYPE_H

#include <iostream>
#include <istream>
#include <fstream>
#include <iterator>
#include <string>
#include <map>
#include <clx/container_wrapper.h>
#include <clx/code_convert.h>
#include <clx/utility.h>
#include <clx/hexdump.h>
#include <cliff/utility.h>
#include <cliff/font/true_type_name_block.h>

namespace cliff {
	/* --------------------------------------------------------------------- */
	/*
	 *  true_type
	 *
	 *  TrueType (*.ttf) 形式のファイルを解析するクラス．
	 *  *.ttf は，様々な情報を“ブロック”と言う単位で管理している．
	 *  ヘッダには，各ブロックのオフセットとブロック長が記載されており，
	 *  その値を基にブロック毎にそれぞれ処理が行われる．
	 */
	/* --------------------------------------------------------------------- */
	class true_type {
	public:
		true_type() :
			pos_(), names_() {}
		
		explicit true_type(istream_type& in) :
			pos_(), names_() {
			this->read(in);
		}
		
		explicit true_type(const string_type& path) :
			pos_(), names_() {
			this->read(path);
		}
		
		~true_type() throw() {}
		
		bool read(istream_type& in) {
			if (!this->read_header(in)) return false;
			
			bool status = true;
			for (std::map<size_type, pos_type>::iterator it = pos_.begin(); it != pos_.end(); ++it) {
				// ToDo: 必要なブロックに関して，処理するためのコードを追加していく．
				switch (it->first) {
				case clx::combine4_<'n', 'a', 'm', 'e'>::value:
					status &= names_.read(in, it->second);
					break;
				default:
					break;
				}
			}
			
			return status;
		}
		
		bool read(const string_type& path) {
			std::basic_ifstream<char_type> ifs(path.c_str());
			if (ifs.fail()) return false;
			return this->read(ifs);
		}
		
		/* ----------------------------------------------------------------- */
		//  access methods.
		/* ----------------------------------------------------------------- */
		name_block& names() { return names_; }
		const name_block& names() const { return names_; }
		
	private:
		bool read_header(istream_type& in) {
			size_type sig = 0;
			if (!clx::get(in, sig, clx::endian::big)) return false;
			if (sig == clx::combine4_<'t', 't', 'c', 'f'>::value) return false;
			
			unsigned short count = 0;
			if (!clx::get(in, count, clx::endian::big)) return false;
			if (!this->check_version(sig, count)) return false;
			in.seekg(6, std::ios_base::cur); // sr, es, rs
			
			for (size_type i = 0; i < count; ++i) {
				size_type tag = 0, checksum = 0, offset = 0, length = 0;
				if (!clx::get(in, tag, clx::endian::big)) return false;
				if (!clx::get(in, checksum, clx::endian::big)) return false;
				if (!clx::get(in, offset, clx::endian::big)) return false;
				if (!clx::get(in, length, clx::endian::big)) return false;
				pos_[tag] = std::make_pair(offset, length);
				//std::cout << format("tag: %x, off: %x, len: %d") % tag % offset % length << std::endl;
			}
			
			return true;
		}
		
		bool check_version(size_type sig, size_type count) {
			if (sig == clx::combine4_<'S', 'p', 'l', 'i'>::value ||
				sig == clx::combine4_<'%', '!', 'P', 'S'>::value ||
				sig == clx::combine4_<'t', 'y', 'p', '1'>::value) {
				return false;
			}
			else if (sig >= 0x80000000 && count == 0) return false;
			else if ((sig >> 24) == 1 && ((sig >> 16) & 0xff) == 0 && ((sig >> 8) & 0xff) == 4) return false; // CFF?
			
			return true; // "OTTO", "true", or others.
		}
		
	private:
		typedef std::pair<size_type, size_type> pos_type;
		std::map<size_type, pos_type> pos_;
		name_block names_;
	};
}

#endif // CLIFF_FONT_TRUE_TYPE_H
