/* ------------------------------------------------------------------------- */
/*
 *  true_type_block.h
 *
 *  Copyright (c) 2009 - 2010, clown. All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    - No names of its contributors may be used to endorse or promote
 *      products derived from this software without specific prior written
 *      permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  Copyright (c) 2000,2001,2002,2003,2004,2005,2006,2007,2008,2009
 *  by George Williams 
 *
 *  The source code is derived from the FontForge Project. (See the
 *  following URL: http://fontforge.sourceforge.net/)
 *
 *  Last-modified: Thu 07 Jan 2010 16:49:00 JST
 */
/* ------------------------------------------------------------------------- */
#ifndef CLIFF_FONT_TRUE_TYPE_NAME_BLOCK_H
#define CLIFF_FONT_TRUE_TYPE_NAME_BLOCK_H

#include <utility>
#include <clx/container_wrapper.h>
#include <clx/utility.h>
#include <cliff/config.h>
#include <cliff/utility.h>

namespace cliff {
	/* --------------------------------------------------------------------- */
	/*
	 *  name_type
	 *
	 *  ある language に対応するフォントに関する各種情報を格納するための
	 *  クラス．platform, encoding の値は，それぞれ下記の通り．
	 *
	 *  platform:
	 *    0: Apple Unicode, 1: Macintosh, 2: ISO, 3: Microsoft
	 *
	 *  encoding:
	 *    0: Symbol, 1: Unicode, 2: ShiftJIS, 3: Big5,
	 *    4: PRC, 5: Wansung, 6: Johab
	 *
	 *  ただし，CLIFF_USE_BABEL が有効の場合，encoding の値に関係なく
	 *  文字列のエンコードは全て UTF-8 である．
	 */
	/* --------------------------------------------------------------------- */
	class name_type {
	public:
		typedef unsigned short value_type;
		
		name_type() :
			platform_(0), encoding_(0),
			copyright_(), family_(), full_(), ps_(), version_() {}
		
		/* ----------------------------------------------------------------- */
		//  access methods.
		/* ----------------------------------------------------------------- */
		const value_type& platform() const { return platform_; }
		const value_type& encoding() const { return encoding_; }
		const string_type& copyright() const { return copyright_; }
		const string_type& familyname() const { return family_; }
		const string_type& fullname() const { return full_; }
		const string_type& psname() const { return ps_; }
		const string_type& version() const { return version_; }
		
		/* ----------------------------------------------------------------- */
		//  access methods.
		/* ----------------------------------------------------------------- */
		void platform(const value_type& cp) { platform_ = cp; }
		void encoding(const value_type& cp) { encoding_ = cp; }
		void copyright(const string_type& cp) { copyright_ = cp; }
		void familyname(const string_type& cp) { family_ = cp; }
		void fullname(const string_type& cp) { full_ = cp; }
		void psname(const string_type& cp) { ps_ = cp; }
		void version(const string_type& cp) { version_ = cp; }
		
	private:
		value_type platform_;
		value_type encoding_;
		string_type copyright_;		// nameid = 0
		string_type family_;		// nameid = 1
		string_type full_;			// nameid = 4
		string_type ps_;			// nameid = 6
		string_type version_;		// nameid = 5
	};
	
	/* --------------------------------------------------------------------- */
	//  name_block
	/* --------------------------------------------------------------------- */
	class name_block : public clx::map_wrapper<size_type, name_type> {
	private:
		typedef clx::map_wrapper<size_type, name_type> super;
	public:
		name_block() : super() {}
		
		virtual ~name_block() throw() {}
		
		bool read(istream_type& in, const std::pair<size_type, size_type>& pos) {
			in.seekg(pos.first);
			if (in.fail()) return false;
			
			unsigned short format = 0, count = 0, taboff = 0;
			if (!clx::get(in, format, clx::endian::big)) return false;
			if (!clx::get(in, count, clx::endian::big)) return false;
			if (!clx::get(in, taboff, clx::endian::big)) return false;
			if (count == 0) return false;
			
			for (size_type i = 0; i < count; ++i) {
				unsigned short platform = 0, encoding = 0, language = 0, id = 0;
				if (!clx::get(in, platform, clx::endian::big)) return false;
				if (!clx::get(in, encoding, clx::endian::big)) return false;
				if (!clx::get(in, language, clx::endian::big)) return false;
				if (!clx::get(in, id, clx::endian::big)) return false;
				
				super::mapped_type& elem = member[static_cast<size_type>(language)];
				if (elem.platform() == 0) elem.platform(platform);
				if (elem.encoding() == 0) elem.encoding(encoding);
				//assert(elem.platform() == platform);
				//assert(elem.encoding() == encoding);
				
				unsigned short offset = 0, length = 0;
				if (!clx::get(in, length, clx::endian::big)) return false;
				if (!clx::get(in, offset, clx::endian::big)) return false;
				
				size_type cur = in.tellg();
				in.seekg(pos.first + taboff + offset);
				if (in.fail()) return false;
				
				string_type s(length, char_type(0));
				in.read(reinterpret_cast<char_type*>(&s.at(0)), length);
				if (in.fail() || in.gcount() < static_cast<int>(length)) return false;
				in.seekg(cur);
				
#ifdef CLIFF_USE_BABEL
				/*
				 * UTF-16 を UTF-8 に変換する（オプションが有効の場合）．
				 *
				 * ToDo:
				 *   platform == 3 (Microsoft) の場合は，encoding に関わらず
				 *   Unicode の模様．それ以外の場合（platform != 3 の場合）を
				 *   調査する．
				 */
				if (platform == 3) s = babel::unicode_to_utf8(clx::widen(s));
#endif
				
				switch (id) {
				case 0: elem.copyright(s);  break;
				case 1: elem.familyname(s); break;
				case 4: elem.fullname(s);   break;
				case 5: elem.version(s);    break;
				case 6: elem.psname(s);     break;
				default:
					//std::cout << "ID: " << id << std::endl;
					//std::cout << "description: " << s << std::endl;
					break;
				}
			}
			
			return true;
		}
	};
}

#endif // CLIFF_FONT_TRUE_TYPE_NAME_BLOCK_H
